@@ -1,21 +1,36 @@
 """Some utilities for handling XML"""
 import StringIO
 import datetime
+import logging
 
 import xml.etree.ElementTree as ET
 
+from .datatypes import DataType
+from healthvaultlib.exceptions import HealthVaultException
+
+
+logger = logging.getLogger(__name__)
+
+
+def pretty_xml(xml):
+    """Given a string with XML, return a string with the XML formatted pretty"""
+    from xml.dom import minidom
+    dom = minidom.parseString(xml)
+    return dom.toprettyxml()
 
 def elt_to_string(elt):
     """Given an ElementTree element object, return a string with nicely formatted XML"""
+    s = elt_as_string(elt)
+    # Now s has the XML as one long string, but that's hard to read.
+    # This is inefficient, but we only use it when debugging.
+    return pretty_xml(s)
+
+def elt_as_string(elt):
     et = ET.ElementTree(elt)
     s = StringIO.StringIO()
     et.write(s)
     s = s.getvalue()
-    # Now s has the XML as one long string, but that's hard to read.
-    # This is inefficient, but we only use it when debugging.
-    from xml.dom import minidom
-    dom = minidom.parseString(s)
-    return dom.toprettyxml()
+    return s
 
 # Little utils to help pull data out of the XML
 def text_list(elt, xpath):
@@ -378,4 +393,197 @@ def parse_awakening(elt):
     return dict(
         when=parse_time(elt.find('when')),
         minutes=int_or_none(elt, 'minutes'),
-    )
\ No newline at end of file
+    )
+
+def parse_subscription(elt):
+    """
+    https://platform.healthvault-ppe.com/platform/XSD/subscription.xsd
+    """
+    return dict(
+        common=parse_subscription_common(elt.find('common')),
+        record_item_changed_event=parse_record_item_changed_event(elt.find('record-item-changed-event')),
+    )
+
+def parse_subscription_common(elt):
+    return dict(
+        id=text_or_none(elt, 'id'),
+        notification_authentication_info=parse_notification_authentication_info(elt.find('notification-authentication-info')),
+        notification_channel=parse_notification_channel(elt.find('notification-channel')),
+    )
+
+def parse_notification_authentication_info(elt):
+    return dict(
+        hv_eventing_shared_key=parse_hv_eventing_shared_key(elt.find('hv-eventing-shared-key')),
+    )
+
+def parse_hv_eventing_shared_key(elt):
+    return dict(
+        notification_key=text_or_none(elt, 'notification-key'),
+        notification_key_version_id=text_or_none(elt, 'notification-key-version-id'),
+    )
+
+def parse_notification_channel(elt):
+    return dict(
+        http_notification_channel=parse_optional_item(elt, 'http-notification-channel', parse_http_notification_channel),
+    )
+
+def parse_http_notification_channel(elt):
+    return dict(
+        url=text_or_none(elt, 'url'),
+    )
+
+def parse_record_item_changed_event(elt):
+    return dict(
+        filters=[parse_record_item_changed_event_filter(f) for f in elt.findall('filters/filter')]
+    )
+
+def parse_record_item_changed_event_filter(elt):
+    type_ids = elt.find('type-ids')
+    return dict(
+        type_ids=[item.text for item in type_ids.findall('type-id')]
+    )
+
+def parse_notification(elt):
+    return dict(
+        common=parse_notification_common(elt.find('common')),
+        record_change_notification=parse_optional_item(elt, 'record-change-notification', parse_record_change_notification),
+    )
+
+def parse_notification_common(elt):
+    return dict(
+        subscription_id=text_or_none(elt, 'subscription-id'),
+    )
+
+def parse_record_change_notification(elt):
+    return dict(
+        person_id=text_or_none(elt, 'person-id'),
+        record_id=text_or_none(elt, 'record-id'),
+        things=[parse_notification_thing(t) for t in elt.findall('things/thing')],
+    )
+
+def parse_notification_thing(elt):
+    return text_or_none(elt, 'thing-id')
+
+def parse_blood_glucose(elt):
+    # http://developer.healthvault.com/pages/types/type.aspx?id=879e7c04-4e8a-4707-9ad3-b054df467ce4
+
+    """
+     <blood-glucose>
+  <when>
+    <date>
+      <y>2006</y>
+      <m>1</m>
+      <d>1</d>
+    </date>
+    <time>
+      <h>9</h>
+      <m>30</m>
+      <s>0</s>
+      <f>0</f>
+    </time>
+  </when>
+  <value>
+    <mmolPerL>7.444444</mmolPerL>
+    <display units="mmolPerL">7.444444</display>
+  </value>
+  <glucose-measurement-type>
+    <text>Whole blood</text>
+    <code>
+      <value>wb</value>
+      <family>wc</family>
+      <type>glucose-measurement-type</type>
+      <version>1</version>
+    </code>
+  </glucose-measurement-type>
+  <outside-operating-temp>true</outside-operating-temp>
+  <is-control-test>true</is-control-test>
+  <normalcy>1</normalcy>
+  <measurement-context>
+    <text>Before meal</text>
+    <code>
+      <value>BeforeMeal</value>
+      <family>wc</family>
+      <type>glucose-measurement-context</type>
+      <version>1</version>
+    </code>
+  </measurement-context>
+</blood-glucose>
+    """
+    return dict(
+        when=when_to_datetime(elt.find('when')),
+        value=parse_blood_glucose_value(elt.find('value')),
+        glucose_measurement_type=parse_codable_value(elt.find('glucose-measurement-type')),
+        outside_operating_temperature=boolean_or_none(elt, 'outside-operating_temp'),
+        is_control_test=boolean_or_none(elt, 'is-control-test'),
+        normalcy=int_or_none(elt, 'normalcy'),
+        measurement_context=parse_optional_item(elt, 'measurement-context', parse_codable_value),
+    )
+
+def parse_blood_glucose_value(elt):
+    # http://developer.healthvault.com/pages/types/type.aspx?id=3e730686-781f-4616-aa0d-817bba8eb141#blood-glucose-value
+    return dict(
+        mmolperl=parse_positive_double(elt.find('mmolPerL')),
+        display=parse_optional_item(elt, 'display', parse_display_value)
+    )
+
+def parse_connect_request(elt):
+    # https://platform.healthvault-ppe.com/platform/XSD/response-getauthorizedconnectrequests.xsd
+    return dict(
+        person_id=text_or_none(elt, 'person-id'),
+        app_specific_record_id=text_or_none('record-id/app-specific-record-id'),
+        app_id=text_or_none(elt, 'app-id'),
+        external_id=text_or_none(elt, 'external-id'),
+    )
+
+def parse_group(group):
+    """Given an element that contains a <group>...</group>
+    return whatever parsing that group as a response to the specific API
+    would have returned.
+
+    :param elementtree group: A `group` element.  Its type is inferred
+    from the <thing><type-id>xxxxxxxx</type-id></thing> value.
+    """
+
+    data_type = group.find('thing/type-id').text
+
+    # FIXME: should we replace all these if/elif clauses with a table-driven implementation?
+    if data_type == DataType.BASIC_DEMOGRAPHIC_DATA:
+        basic = group.find('thing/data-xml/basic')
+        return dict(
+            gender=text_or_none(basic, 'gender'),
+            birthyear=int_or_none(basic, 'birthyear'),
+            country_text=text_or_none(basic, 'country/text'),
+            country_code=text_or_none(basic, 'country/code/value'),
+            postcode=text_or_none(basic, 'postcode'),
+            state=text_or_none(basic,'state/text')
+        )
+    elif data_type == DataType.BLOOD_GLUCOSE_MEASUREMENT:
+        return [parse_blood_glucose(item) for item in group.findall('thing/data-xml/blood-glucose')]
+    elif data_type == DataType.BLOOD_PRESSURE_MEASUREMENTS:
+        things = []
+        for thing in group.findall('thing/data-xml/blood-pressure'):
+            things.append(dict(
+                when = when_to_datetime(thing.find("when")),
+                systolic = int_or_none(thing, 'systolic'),
+                diastolic = int_or_none(thing, 'diastolic'),
+                pulse = int_or_none(thing, 'pulse'),
+                irregular_heartbeat = boolean_or_none(thing, 'irregular-heartbeat'),
+            ))
+        return things
+    elif data_type == DataType.DEVICES:
+        return [parse_device(e) for e in group.findall('thing/data-xml/device')]
+    elif data_type == DataType.EXERCISE:
+        return [parse_exercise(e) for e in group.findall('thing/data-xml/exercise')]
+    elif data_type == DataType.HEIGHT_MEASUREMENTS:
+        return [parse_height(e) for e in group.findall('thing/data-xml/height')]
+    elif data_type == DataType.SLEEP_SESSIONS:
+        return [parse_sleep_session(s) for s in group.findall('thing/data-xml/sleep-am')]
+    elif data_type == DataType.WEIGHT_MEASUREMENTS:
+        return [parse_weight(e) for e in group.findall('thing/data-xml/weight')]
+    else:
+        # import here to avoid circular imports
+        raise HealthVaultException("Unknown data type in group response: name='%s'" % data_type)
+
+
+
+
